import sys
try:
    import StringIO
    import cStringIO
except ImportError:
    from io import BytesIO

running_python_3 = sys.version_info[0] == 3


def range_fn(*args):
    if running_python_3:
        return range(*args)
    else:
        return xrange(*args)


def is_string_io(instance):
    if running_python_3:
       return isinstance(instance, BytesIO)
    else:
        return isinstance(instance, (StringIO.StringIO,
                                     cStringIO.InputType,
                                     cStringIO.OutputType))



import os
from base64 import b64encode
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend


#######
# IOC #
#######
def get_types(ioc):
    return '||'.join(attr_type for attr_type in ioc)

def get_values(ioc):
    return '||'.join(ioc[attr_type] for attr_type in ioc)

def get_types_values(ioc):
    return (get_types(ioc), get_values(ioc))


#######
# AES #
#######
def aes_create_rule(key, message, attr_types, salt):
    if len(key) < 32:
        print('Key length is not enough, It is adviced to change the algorithm used!')
    nonce = os.urandom(16)

    # Encrypt
    backend = default_backend()
    cipher = Cipher(algorithms.AES(key), modes.CTR(nonce), backend=backend)
    encryptor = cipher.encryptor()
    ct_check = encryptor.update(b'\x00'*16)
    ct_message = encryptor.update(message.encode('utf-8'))
    ct_message += encryptor.finalize()

    # Create the rule
    rule = {}
    rule['salt'] = b64encode(salt).decode('ascii')
    rule['attributes'] = attr_types
    rule['nonce'] = b64encode(nonce).decode('ascii')
    rule['ciphertext-check'] = b64encode(ct_check).decode('ascii')
    rule['ciphertext'] = b64encode(ct_message).decode('ascii')

    return rule

def aes_match_rule(key, password, nonce, ciphertext):
    backend = default_backend()
    cipher = Cipher(algorithms.AES(key), modes.CTR(nonce), backend=backend)
    dec = cipher.decryptor()
    # A match is found when the first block is all null bytes
    if dec.update(ciphertext[0]) == b'\x00'*16:
        plaintext = dec.update(ciphertext[1]) + dec.finalize()
        return (True, plaintext)
    else:
        return (False, '')




"""
Cryptographic system with bcrypt
"""
from pypraware_crypto.interface import Crypto
from pypraware_crypto.helper import *
import configparser
import glob, hashlib, os
from base64 import b64encode

# hash and crypto import
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
import bcrypt

class Bcrypt(Crypto):
    def __init__(self, conf, metadata=None):
        self.conf = conf
        self.btoken = bytes(conf['misp']['token'], encoding='ascii')
        self.round = int(self.conf['bcrypt']['round'])
        self.ipround = int(self.conf['bcrypt']['ipround'])
        # For matching (only token is kept from config file)
        if metadata is not None:
            metadata = metadata['crypto']
            self.round = int(metadata['round'])
            self.ipround = int(metadata['ipround'])

    def derive_key(self, bpassword, bsalt, attr_types):
        """
        Generate the key used for encryption
        Bcrypt truncates password to 72 bytes. 
            password + token : For long data, token will not be included
            token + password : only use 32 bytes of data
            => Hash (password + token) as the password
        """
        rd = 1
        if attr_types in ["ip-dst", "ip-src", "ip-src||port", "ip-dst||port"]:
            rd = self.ipround
        else:
            rd = self.round
        # Solve the truncation problem
        digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
        digest.update(bpassword)
        digest.update(self.btoken)
        token_pass = digest.finalize()
        key =  bcrypt.kdf(password = token_pass, 
                salt = bsalt,
                desired_key_bytes = 32,
                rounds = rd)
        return key

    def create_rule(self, ioc, message):
        salt = os.urandom(16)
        attr_types, password = get_types_values(ioc)
        dk = self.derive_key(password.encode('utf8'), salt, attr_types)
        return aes_create_rule(dk, message, attr_types, salt)

    def match(self, attributes, rule, queue):
        """
        Sometimes we don't need to decrypt the whole
        ciphertext to know if there is a match
        as it is the case here thanks to ctr mode
        """
        rule_attr = rule['attributes']
        match = False
        try:
            password = '||'.join([attributes[attr] for attr in rule_attr])
            attr_types = '||'.join(attr_type for attr_type in rule_attr)
        except:
            pass # Nothing to do
        dk = self.derive_key(password.encode('utf8'), rule['salt'], attr_types)
        ciphertext = [rule['ciphertext-check'], rule['ciphertext']]
        match, plaintext = aes_match_rule(dk, password, rule['nonce'],\
                ciphertext)

        if match:
            queue.put("IOC matched for: {}\nSecret Message (uuid-event id-date)\n===================================\n{}\n".format(attributes, plaintext.decode('utf-8')))



    def save_meta(self):
        meta = configparser.ConfigParser()
        meta['crypto'] = {}
        meta['crypto']['name'] = 'bcrypt' 
        meta['crypto']['round'] = str(self.round)
        meta['crypto']['ipround'] = str(self.ipround)
        with open(self.conf['rules']['location'] + '/metadata', 'w') as config:
            meta.write(config)




"""
Cryptographic system with HKDF: HMAC-based Extract-and-Expand Key Derivation Function
Should be used only if the size of the domain of the type of data
is enough to avoid bruteforce attacks
"""
from pypraware_crypto.interface import Crypto
from pypraware_crypto.helper import *
import configparser
import glob, hashlib, os
from base64 import b64encode

# hash and crypto import
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.hkdf import HKDF

class HKDF(Crypto):
    def __init__(self, conf, metadata=None):
        self.conf = conf
        self.btoken = bytes(conf['misp']['token'], encoding='ascii')

        # For matching (only token is kept from config file)
        if metadata is not None:
            metadata = metadata['crypto']

    def derive_key(self, bpassword, bsalt, attr_types):
        """
        Generate the key used for encryption
        """
        passToken = bpassword + self.btoken
        hkdf = HKDF(
            algorithm=hashes.SHA256(),
            length=32,
            salt=bsalt,
            info=self.btoken,
            backend=default_backend())

        return hkdf.derive(bpassword)

    def create_rule(self, ioc, message):
        salt = os.urandom(16)
        attr_types, password = get_types_values(ioc)
        dk = self.derive_key(password.encode('utf8'), salt, attr_types)
        return aes_create_rule(dk, message, attr_types, salt)

    def match(self, attributes, rule, queue):
        rule_attr = rule['attributes']
        match = False
        try:
            password = '||'.join([attributes[attr] for attr in rule_attr])
            attr_types = '||'.join(attr_type for attr_type in rule_attr)
        except:
            pass # Nothing to do
        dk = self.derive_key(password.encode('utf8'), rule['salt'], attr_types)
        ciphertext = [rule['ciphertext-check'], rule['ciphertext']]
        match, plaintext = aes_match_rule(dk, password, rule['nonce'],\
                ciphertext)

        if match:
            queue.put("IOC matched for: {}\nSecret Message (uuid-event id-date)\n===================================\n{}\n".format(attributes, plaintext.decode('utf-8')))



    def save_meta(self):
        meta = configparser.ConfigParser()
        meta['crypto'] = {}
        meta['crypto']['name'] = 'hkdf'
        with open(self.conf['rules']['location'] + '/metadata', 'w') as config:
            meta.write(config)




"""
Cryptographic system for bloom filter
"""
from pypraware_crypto.interface import Crypto
import configparser
import glob, hashlib, os
from base64 import b64encode

# Hash and crypto import
from pypraware_crypto.pybloom import BloomFilter

class Bloom_filter(Crypto):
    def __init__(self, conf, metadata=None, rate=None):
        self.conf = conf
        self.token = conf['misp']['token']
        self.passwords = list()
        self.rate = rate
        if not rate:
            self.rate = conf['bloom_filter']['error_rate']
        # If for matching
        if metadata != None:
            filename = self.conf['rules']['location'] + '/joker'
            with open(filename, 'rb') as fd:
                self.f = BloomFilter.fromfile(fd)

    def create_rule(self, ioc, message):
        """
        We need to create one rule, thus we need a state
        """
        if (len(ioc)>1):
            # We also add the concatenation of the two values
            long_pass = '||'.join([ioc[attr] for attr in ioc])
            self.passwords.append(long_pass + self.token)
        
        for attr in ioc:
            self.passwords.append(ioc[attr]+ self.token)

        return {'joker':True}


    def check(self, attributes, rule):
        """
        Return a list of password to test or an empty list
        """
        passwords = list()

        # Only exists types with two elements ( redis matching
        # has more attributes and it is not usefull to use
        # long pass )
        if (len(attributes)==2):
            # We also add the concatenation of the two values
            long_pass = '||'.join([attributes[attr] for attr in attributes])
            passwords.append(long_pass + self.token)

        for attr in attributes:
            passwords.append(attributes[attr]+ self.token)

        matchPasswords = list()
        for p in passwords:
            if p in self.f:
                matchPasswords.append(p)

        return matchPasswords
        

    def match(self, attributes, rule, queue):
        for p in self.check(attributes, rule):
            queue.put("Value(s) {} matched for {}\n".format(attributes, p[:-len(self.token)]))


    def write_bloom(self):
        # Create Bloom filter
        f = BloomFilter(capacity=len(self.passwords), error_rate=float(self.rate))
        [f.add(password) for password in self.passwords ]
        with open(self.conf['rules']['location'] + '/joker', 'wb') as fd:
            f.tofile(fd)

    def save_meta(self):
        meta = configparser.ConfigParser()
        meta['crypto'] = {}
        meta['crypto']['name'] = 'bloom_filter' 
        err_rate = self.rate
        meta['crypto']['error_rate'] = err_rate
        with open(self.conf['rules']['location'] + '/metadata', 'w') as config:
            meta.write(config)
        
        self.write_bloom()

